import pyarrow
import rpy2.rinterface as rinterface
import rpy2.robjects as robjects
import rpy2.robjects.conversion as conversion
import rpy2.robjects.packages as packages
import typing
import warnings
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    import rpy2.robjects

rarrow = packages.importr('arrow')
TARGET_VERSION = '2.0.'
if not rarrow.__version__.startswith(TARGET_VERSION):
    warnings.warn(
        'This was designed againt arrow versions starting with %s'
        ' but you have %s' %
        (TARGET_VERSION, rarrow.__version__))


def pyarrow_to_r_array(
        obj: 'pyarrow.lilb.Array'
):
    """Create an R `arrow::Arrow` object from a pyarrow Array.

    This is sharing the C/C++ object between the two languages.
    The returned object depends on the active conversion rule in
    rpy2. By default it will be an `rpy2.robjects.Environment`.
    """
    schema_ptr = rarrow.allocate_arrow_schema()[0]
    array_ptr = rarrow.allocate_arrow_array()[0]
    try:
        obj._export_to_c(int(array_ptr), int(schema_ptr))
        r_array = rarrow.ImportArray(array_ptr, schema_ptr)
    finally:
        rarrow.delete_arrow_schema(schema_ptr)
        rarrow.delete_arrow_array(array_ptr)
    return r_array


def rarrow_to_py_array(
        obj: 'rpy2.robjects.Environment'
):
    """Create a pyarrow array from an R `arrow::Array` object.

    This is sharing the C/C++ object between the two languages.
    """

    schema_ptr = rarrow.allocate_arrow_schema()[0]
    array_ptr = rarrow.allocate_arrow_array()[0]
    try:
        rarrow.ExportArray(obj, array_ptr, schema_ptr)
        py_array = pyarrow.Array._import_from_c(int(array_ptr),
                                                int(schema_ptr))
    finally:
        rarrow.delete_arrow_schema(schema_ptr)
        rarrow.delete_arrow_array(array_ptr)
    return py_array


def pyarrow_to_r_recordbatch(
        obj: 'pyarrow.lib.RecordBatch'
):
    """Create an R `arrow::RecordBatch` object from a pyarrow RecordBatch.

    This is sharing the C/C++ object between the two languages.
    The returned object depends on the active conversion rule in
    rpy2. By default it will be an `rpy2.robjects.Environment`.
    """

    schema_ptr = rarrow.allocate_arrow_schema()[0]
    array_ptr = rarrow.allocate_arrow_array()[0]
    try:
        obj._export_to_c(int(array_ptr), int(schema_ptr))
        r_recordbatch = rarrow.ImportRecordBatch(array_ptr, schema_ptr)
    finally:
        rarrow.delete_arrow_schema(schema_ptr)
        rarrow.delete_arrow_array(array_ptr)
    return r_recordbatch


def pyarrow_to_r_chunkedarray(
        obj: 'pyarrow.lib.ChunkedArray'
):
    """Create an R `arrow::ChunkedArray` object from a pyarrow ChunkedArray.

    This is sharing the C/C++ object between the two languages.
    The returned object depends on the active conversion rule in
    rpy2. By default it will be an `rpy2.robjects.Environment`.
    """
    chunks = tuple(pyarrow_to_r_array(x) for x in obj.chunks)
    print(len(chunks))
    import time
    t0 = time.time()
    res = rarrow.ChunkedArray['create'](*chunks)
    t1 = time.time()
    print(f'rarrow.ChunkedArray: {t1-t0}')
    return res


def rarrow_to_py_chunkedarray(
        obj: 'rpy2.robjects.Environment'
):
    """Create a pyarrow chunked array from an R `arrow::ChunkedArray` object.

    This is sharing the C/C++ object between the two languages.
    """
    chunks = tuple(rarrow_to_py_array(x) for x in obj['chunks'])
    return pyarrow.chunked_array(chunks)


def pyarrow_to_r_table(
        obj: 'pyarrow.lib.Table',
        py2rpy: typing.Optional[
            conversion.Converter] = None
):
    """Create an R `arrow::Table` object from a pyarrow Table.

    This is sharing the C/C++ object between the two languages.
    The returned object depends on the active conversion rule in
    rpy2. By default it will be an `rpy2.robjects.Environment`.
    """

    if py2rpy is None:
        py2rpy = converter
    # TODO: this is using the converter defined in the module,
    # not the converter currently in rpy2.robjects.conversion.
    kwargs = dict(
        (k, converter.py2rpy(v))
        for k, v in zip(obj.schema.names, obj.columns)
    )
    kwargs['schema'] = pyarrow_to_r_schema(obj.schema)
    return rarrow.Table['create'](**kwargs)


def rarrow_to_py_table(
        obj: 'rpy2.robjects.Environment',
        rpy2py: typing.Optional[
            conversion.Converter] = None
):
    """Create a pyarrow Table fomr an R `arrow::Table` object.

    This is sharing the C/C++ object between the two languages.
    The returned object depends on the active conversion rule in
    rpy2.
    """

    if rpy2py is None:
        rpy2py = converter
    # TODO: rpy2 conversion forces something a little kludgy here.
    columns = [
        (rpy2py._rpy2py_nc_map[rinterface.SexpEnvironment][x.rclass[0]](x))
        for x in obj['columns']
    ]
    schema = rarrow_to_py_schema(obj['schema'])
    py_table = pyarrow.Table.from_arrays(columns,
                                         schema=schema)
    return py_table


def pyarrow_to_r_schema(
        obj: 'pyarrow.lib.Schema'
):
    """Create an R `arrow::Schema` object from a pyarrow Schema.

    This is sharing the C/C++ object between the two languages.
    The returned object depends on the active conversion rule in
    rpy2. By default it will be an `rpy2.robjects.Environment`.
    """

    schema_ptr = rarrow.allocate_arrow_schema()[0]
    try:
        obj._export_to_c(int(schema_ptr))
        r_schema = rarrow.ImportSchema(schema_ptr)
    finally:
        rarrow.delete_arrow_schema(schema_ptr)
    return r_schema


def rarrow_to_py_schema(
        obj: 'rpy2.robjects.Environment'
):
    """Create a pyarrow Schema fomr an R `arrow::Schema` object.

    This is sharing the C/C++ object between the two languages.
    The returned object depends on the active conversion rule in
    rpy2. By default it will be an `rpy2.robjects.Environment`.
    """

    schema_ptr = rarrow.allocate_arrow_schema()[0]
    try:
        rarrow.ExportSchema(obj, schema_ptr)
        py_schema = pyarrow.Schema._import_from_c(int(schema_ptr))
    finally:
        rarrow.delete_arrow_schema(schema_ptr)
    return py_schema


converter = conversion.Converter('default arrow conversion',
                                 template=robjects.default_converter)
# Pyarrow to R arrow.
converter.py2rpy.register(pyarrow.Array, pyarrow_to_r_array)
converter.py2rpy.register(pyarrow.ChunkedArray,
                          pyarrow_to_r_chunkedarray)
converter.py2rpy.register(pyarrow.Schema, pyarrow_to_r_schema)
converter.py2rpy.register(pyarrow.Table, pyarrow_to_r_table)
# R arrow to pyarrow.
converter._rpy2py_nc_map.update(
    {
        rinterface.SexpEnvironment:
        conversion.NameClassMap(robjects.Environment)
    }
)

# TODO: use complete class name hierarchy to be safer?
converter._rpy2py_nc_map[rinterface.SexpEnvironment].update(
    {
        'Array': rarrow_to_py_array,
        'ChunkedArray': rarrow_to_py_chunkedarray,
        'Schema': rarrow_to_py_schema,
        'Table': rarrow_to_py_table
    }
)



from rpy2.robjects.packages import importr
from rpy2 import robjects
import rpy2_R6.r6b as r6b
import rpy2.rinterface_lib.callbacks

r_arrow = importr('arrow')
methods = importr('methods')


class CustomR6(r6b.R6):

    def __init__(self, obj):
        super().__init__(obj)
        self.ro = robjects.vectors.VectorOperationsDelegator(self)
        self.rx = robjects.vectors.ExtractDelegator(self)
        self.rx2 = robjects.vectors.DoubleExtractDelegator(self)

    def __str__(self):
        s = []
        with (rpy2.rinterface_lib
              .callbacks.obj_in_module(rpy2.rinterface_lib.callbacks,
                                       'consolewrite_print', s.append)):
            methods.show(self)
        s = str.join('', s)
        return s



class ArrowClassGenerator(r6b.R6DynamicClassGenerator):

    __DEFAULT_ATTRS__ = r6b.R6DynamicClassGenerator.__DEFAULT_ATTRS__.copy()
    __DEFAULT_ATTRS__.update({'new': r6b.r6_factorymethod,
                              'create': r6b.r6_factorymethod})
    __CLASSMAP__ = lambda x: r6b._dynamic_classmap(x, bases = (CustomR6, ))


ArrayGen = ArrowClassGenerator(r_arrow.Array)
TableGen = ArrowClassGenerator(r_arrow.Table)




import array
import pandas
import rpy2.rinterface as rinterface
import rpy2_arrow.pyarrow_rarrow as pyr
import pyarrow


def test_py2r_Array():
    a = array.array('i', [1, 2, 3])
    py_ar = pyarrow.array(a)
    r_ar = pyr.pyarrow_to_r_array(py_ar)
    assert isinstance(r_ar, rinterface.SexpEnvironment)


def test_r2py_Array():
    r_ar = rinterface.evalr("""
    require('arrow')
    arrow::Array$create(c(1, 2, 3))
    """)
    py_ar = pyr.rarrow_to_py_array(r_ar)
    assert isinstance(py_ar, pyarrow.Array)


def test_RecordBatch():
    dataf = pandas.DataFrame.from_dict(
        {'a': [1, 2, 3],
         'b': [4, 5, 6]}
    )
    py_rb = pyarrow.record_batch(dataf)
    r_rb = pyr.pyarrow_to_r_recordbatch(py_rb)
    assert isinstance(r_rb, rinterface.SexpEnvironment)


def test_ChunkedArray():
    py_ca = pyarrow.chunked_array([[1, 2, 3], [4, 5]])
    r_ca = pyr.pyarrow_to_r_chunkedarray(py_ca)
    assert isinstance(r_ca, rinterface.SexpEnvironment)


def test_r2py_ChunkedArray():
    r_ca = rinterface.evalr("""
    require('arrow')
    arrow::ChunkedArray$create(c(1, 2, 3))
    """)
    py_ca = pyr.rarrow_to_py_chunkedarray(r_ca)
    assert isinstance(py_ca, pyarrow.ChunkedArray)


def test_py2r_Table():
    dataf = pandas.DataFrame.from_dict(
        {'a': [1, 2, 3],
         'b': [4, 5, 6]}
    )
    py_tb = pyarrow.record_batch(dataf)
    r_tb = pyr.pyarrow_to_r_table(py_tb)
    assert isinstance(r_tb, rinterface.SexpEnvironment)


def test_r2py_Table():
    r_tb = rinterface.evalr("""
    require('arrow')
    arrow::Table$create(
      data.frame(a = c(1, 2, 3),
                 b = c(4, 5, 6))
    )
    """)
    py_tb = pyr.rarrow_to_py_table(r_tb)
    assert isinstance(py_tb, pyarrow.Table)



from setuptools import setup

pack_version = __import__('rpy2_arrow').__version__

if __name__ == '__main__':

    extras_require = {'R6': ['rpy2_R6'],
                      'test': ['pytest', 'pandas']}
    extras_require['all'] = list(
        set(x for lst in extras_require.values()
            for x in lst)
    )
    setup(
        name='rpy2-arrow',
        version=pack_version,
        description='Bridge Arrow between Python and R when using rpy2',
        license='MIT',
        requires=['pyarrow', 'rpy2'],
        extras_require=extras_require,
        packages=['rpy2_arrow', ],
        url='https://github.com/rpy2/rpy2-arrow',
        classifiers=['Programming Language :: Python',
                     'Programming Language :: Python :: 3',
                     'Programming Language :: Python :: 3.6',
                     'Programming Language :: Python :: 3.7',
                     'Programming Language :: Python :: 3.8',
                     ('License :: OSI Approved :: MIT License'),
                     'Intended Audience :: Developers',
                     'Intended Audience :: Science/Research']
    )
